@@ -2,13 +2,10 @@
 
 #include <qpdf/QPDF.hh>
 
-#include <algorithm>
 #include <atomic>
-#include <cstdlib>
 #include <cstring>
 #include <limits>
 #include <map>
-#include <memory.h>
 #include <regex>
 #include <sstream>
 #include <vector>
@@ -17,8 +14,6 @@
 #include <qpdf/FileInputSource.hh>
 #include <qpdf/OffsetInputSource.hh>
 #include <qpdf/Pipeline.hh>
-#include <qpdf/Pl_Discard.hh>
-#include <qpdf/Pl_OStream.hh>
 #include <qpdf/QPDFExc.hh>
 #include <qpdf/QPDFLogger.hh>
 #include <qpdf/QPDFObject_private.hh>
@@ -579,6 +574,39 @@ QPDF::reconstruct_xref(QPDFExc& e)
     }
     m->deleted_objects.clear();
 
+    if (!m->trailer.isInitialized()) {
+        qpdf_offset_t max_offset{0};
+        // If there are any xref streams, take the last one to appear.
+        for (auto const& iter: m->xref_table) {
+            auto entry = iter.second;
+            if (entry.getType() != 1) {
+                continue;
+            }
+            auto oh = getObjectByObjGen(iter.first);
+            try {
+                if (!oh.isStreamOfType("/XRef")) {
+                    continue;
+                }
+            } catch (std::exception&) {
+                continue;
+            }
+            auto offset = entry.getOffset();
+            if (offset > max_offset) {
+                max_offset = offset;
+                setTrailer(oh.getDict());
+            }
+        }
+        if (max_offset > 0) {
+            try {
+                read_xref(max_offset);
+            } catch (std::exception&) {
+                throw damagedPDF(
+                    "", 0, "error decoding candidate xref stream while recovering damaged file");
+            }
+            QTC::TC("qpdf", "QPDF recover xref stream");
+        }
+    }
+
     if (!m->trailer.isInitialized()) {
         // We could check the last encountered object to see if it was an xref stream.  If so, we
         // could try to get the trailer from there.  This may make it possible to recover files with
@@ -678,6 +706,14 @@ QPDF::read_xref(qpdf_offset_t xref_offset)
     // We no longer need the deleted_objects table, so go ahead and clear it out to make sure we
     // never depend on its being set.
     m->deleted_objects.clear();
+
+    // Make sure we keep only the highest generation for any object.
+    QPDFObjGen last_og{-1, 0};
+    for (auto const& og: m->xref_table) {
+        if (og.first.getObj() == last_og.getObj())
+            removeObject(last_og);
+        last_og = og.first;
+    }
 }
 
 bool
@@ -1950,6 +1986,18 @@ QPDF::replaceObject(QPDFObjGen const& og, QPDFObjectHandle oh)
     updateCache(og, oh.getObj(), -1, -1);
 }
 
+void
+QPDF::removeObject(QPDFObjGen og)
+{
+    m->xref_table.erase(og);
+    if (auto cached = m->obj_cache.find(og); cached != m->obj_cache.end()) {
+        // Take care of any object handles that may be floating around.
+        cached->second.object->assign(QPDF_Null::create());
+        cached->second.object->setObjGen(nullptr, QPDFObjGen());
+        m->obj_cache.erase(cached);
+    }
+}
+
 void
 QPDF::replaceReserved(QPDFObjectHandle reserved, QPDFObjectHandle replacement)
 {
@@ -2348,19 +2396,37 @@ QPDF::getCompressibleObjGens()
     QPDFObjectHandle encryption_dict = m->trailer.getKey("/Encrypt");
     QPDFObjGen encryption_dict_og = encryption_dict.getObjGen();
 
-    QPDFObjGen::set visited;
-    std::list<QPDFObjectHandle> queue;
-    queue.push_front(m->trailer);
+    const size_t max_obj = getObjectCount();
+    std::vector<bool> visited(max_obj, false);
+    std::vector<QPDFObjectHandle> queue;
+    queue.reserve(512);
+    queue.push_back(m->trailer);
     std::vector<QPDFObjGen> result;
     while (!queue.empty()) {
-        QPDFObjectHandle obj = queue.front();
-        queue.pop_front();
+        auto obj = queue.back();
+        queue.pop_back();
         if (obj.isIndirect()) {
             QPDFObjGen og = obj.getObjGen();
-            if (!visited.add(og)) {
+            const size_t id = toS(og.getObj() - 1);
+            if (id >= max_obj)
+                throw std::logic_error(
+                    "unexpected object id encountered in getCompressibleObjGens");
+            if (visited[id]) {
                 QTC::TC("qpdf", "QPDF loop detected traversing objects");
                 continue;
             }
+
+            // Check whether this is the current object. If not, remove it (which changes it into a
+            // direct null and therefore stops us from revisiting it) and move on to the next object
+            // in the queue.
+            auto upper = m->obj_cache.upper_bound(og);
+            if (upper != m->obj_cache.end() && upper->first.getObj() == og.getObj()) {
+                removeObject(og);
+                continue;
+            }
+
+            visited[id] = true;
+
             if (og == encryption_dict_og) {
                 QTC::TC("qpdf", "QPDF exclude encryption dictionary");
             } else if (!(obj.isStream() ||
@@ -2381,18 +2447,18 @@ QPDF::getCompressibleObjGens()
                         QTC::TC("qpdf", "QPDF exclude indirect length");
                     }
                 } else {
-                    queue.push_front(value);
+                    queue.push_back(value);
                 }
             }
         } else if (obj.isDictionary()) {
             std::set<std::string> keys = obj.getKeys();
             for (auto iter = keys.rbegin(); iter != keys.rend(); ++iter) {
-                queue.push_front(obj.getKey(*iter));
+                queue.push_back(obj.getKey(*iter));
             }
         } else if (obj.isArray()) {
             int n = obj.getArrayNItems();
             for (int i = 1; i <= n; ++i) {
-                queue.push_front(obj.getArrayItem(n - i));
+                queue.push_back(obj.getArrayItem(n - i));
             }
         }
     }
